@@ -1,8 +1,10 @@
+import asyncio
 import json
 from urllib import response
 import requests
 import logging
 import aiohttp
+import time
 
 from .exceptions import *
 
@@ -11,6 +13,8 @@ logger = logging.getLogger(__name__)
 
 API_ENDPOINT = "https://apigw.verymobile.it/api"
 REQUIRED_HEADERS = {"X-Wind-Client": "app-and", "X-Language": "en", "X-Brand": "DEA"}
+RATE_LIMIT = 7
+MAX_RETRY = 3
 
 class W3API():
     def __init__(self, username, password, session=None) -> None:
@@ -18,10 +22,11 @@ class W3API():
         self._lines = {} # format "LINE_NUMBER": "CONTRACT_ID"
         self._username = username
         self._password = password
+        self._last_request = 0
 
         self._session = aiohttp.ClientSession() if session is None else session
     
-    async def login(self):
+    async def login(self, depth=0):
         
         data = {
             "username": self._username,
@@ -29,11 +34,23 @@ class W3API():
             "rememberMe": False
         }
 
+        if abs(time.time() - self._last_request) < RATE_LIMIT:
+            logger.warn("W3: rate limiting")
+            await asyncio.sleep(RATE_LIMIT)
+
         async with self._session.post(url=f"{API_ENDPOINT}/v4/login/credentials", headers=REQUIRED_HEADERS, json=data) as resp:
+            self._last_request = time.time()
             logger.debug(f"W3 response status {resp.status}")
-            logger.debug(await resp.text())
+ 
+            if resp.status == 429:
+                if depth > MAX_RETRY:
+                    raise RateLimitException()
+
+                logger.warn("W3: rate limiting")
+                await asyncio.sleep(RATE_LIMIT)
+
+                return await self.login(depth+1)
 
-            
             if resp.status != 200:
                 logger.error(f"VeryAPI: response code {resp.status}")
                 raise AuthenticationException(await resp.json())
@@ -57,7 +74,7 @@ class W3API():
     def get_line_numbers(self):
         return self._lines.keys()
 
-    async def _request_unfolded(self, lineid, contractid):
+    async def _request_unfolded(self, lineid, contractid, depth=0):
         if self._token is None:
             self.login()
 
@@ -66,8 +83,22 @@ class W3API():
 
         data = {"contractId": contractid, "lineId": lineid}
 
-        async with self._session.get(url=f"{API_ENDPOINT}/ob/v2/contract/lineunfolded", headers=headers, params=data) as resp:
+        if abs(time.time() - self._last_request) < RATE_LIMIT:
+            logger.warn("W3: rate limiting")
+            await asyncio.sleep(RATE_LIMIT)
 
+        async with self._session.get(url=f"{API_ENDPOINT}/ob/v2/contract/lineunfolded", headers=headers, params=data) as resp:
+            self._last_request = time.time()
+
+            if resp.status == 429:
+                if depth > MAX_RETRY:
+                    raise RateLimitException()
+
+                logger.warn("W3: rate limiting")
+                await asyncio.sleep(RATE_LIMIT)
+
+                return await self._request_unfolded(lineid, contractid, depth+1)
+            
             if resp.status != 200:
                 logger.error(f"VeryAPI: response code {resp.status}")
                 raise RuntimeError(f"VeryAPI: response code {resp.status}")
